import uuid
import time
import re


# ---------- Identity keys
IDENT_KEY_EMAIL="$email"
IDENT_KEY_SMS="$sms"
IDENT_KEY_ANDROIDPUSH="$androidpush"
IDENT_KEY_IOSPUSH="$iospush"
IDENT_KEY_WHATSAPP="$whatsapp"
IDENT_KEY_WEBPUSH="$webpush"
IDENT_KEY_SLACK="$slack"

IDENT_KEYS_ALL = [IDENT_KEY_EMAIL, IDENT_KEY_SMS, IDENT_KEY_ANDROIDPUSH, IDENT_KEY_IOSPUSH,
                  IDENT_KEY_WHATSAPP, IDENT_KEY_WEBPUSH, IDENT_KEY_SLACK]

KEY_PUSHVENDOR = "$pushvendor"

OTHER_RESERVED_KEYS = [
    "$messenger", "$inbox",
    KEY_PUSHVENDOR, "$device_id",
    "$insert_id", "$time",
    "$set", "$set_once", "$add", "$append", "$remove", "$unset",
    "$identify", "$anon_id", "$identified_id",
    "$notification_delivered", "$notification_dismiss", "$notification_clicked",
]

SUPER_PROPERTY_KEYS = [
    "$app_version_string", "$app_build_number", "$brand", "$carrier", "$manufacturer", "$model", "$os",
    "$ss_sdk_version", "$insert_id", "$time"
]

ALL_RESERVED_KEYS = SUPER_PROPERTY_KEYS + OTHER_RESERVED_KEYS + IDENT_KEYS_ALL

# ---------
MOBILE_REGEX = r'^\+[0-9\s]+'
mobile_regex_compiled = re.compile(MOBILE_REGEX)

EMAIL_REGEX = r'^\S+@\S+\.\S+$'
email_regex_compiled = re.compile(EMAIL_REGEX)
# ---------


class _SubscriberInternalHelper:
    """
    Internal helper class
    """
    def __init__(self, distinct_id, workspace_key):
        self.distinct_id = distinct_id
        self.workspace_key = workspace_key
        #
        self.__dict_append = {}
        self.__append_count = 0
        #
        self.__dict_remove = {}
        self.__remove_count = 0
        #
        self.__list_unset = []
        self.__unset_count = 0
        #
        self.__errors = []
        self.__info = []

    def reset(self):
        self.__dict_append, self.__dict_remove, self.__list_unset = {}, {}, []
        self.__append_count, self.__remove_count, self.__unset_count = 0, 0, 0
        self.__errors = []
        self.__info = []

    def get_identity_event(self):
        evt = self.__form_event()
        ret_val = {
            "errors": self.__errors,
            "info": self.__info,
            "event": evt,
            "append": self.__append_count,
            "remove": self.__remove_count,
            "unset": self.__unset_count,
        }
        self.reset()
        return ret_val

    def __form_event(self):
        if (self.__dict_append or self.__dict_remove or self.__list_unset):
            event = {
                "$insert_id": str(uuid.uuid4()).lower(),
                "$time": int(time.time() * 1000),
                "env": self.workspace_key,
                "distinct_id": self.distinct_id,
            }
            if self.__dict_append:
                event["$append"] = self.__dict_append
                self.__append_count += 1
            if self.__dict_remove:
                event["$remove"] = self.__dict_remove
                self.__remove_count += 1
            if self.__list_unset:
                event["$unset"] = self.__list_unset
                self.__unset_count += 1
            return event
        else:
            return None

    # ------------------------
    def __validate_key_basic(self, key, caller):
        if not isinstance(key, (str,)):
            self.__info.append(f"[{caller}] skipping key: {key}. key must be a string")
            return key, False
        key = key.strip()
        if not key:
            self.__info.append(f"[{caller}] skipping key: empty string")
            return key, False
        # -----
        return key, True

    def __validate_key_prefix(self, key, caller):
        if key not in ALL_RESERVED_KEYS:
            prefix_3_chars = key[:3].lower()
            if prefix_3_chars.startswith("$") or prefix_3_chars == "ss_":
                self.__info.append(f"[{caller}] skipping key: {key}. key starting with [$,ss_] are reserved")
                return False
        # ----
        return True

    def __is_identity_key(self, key):
        return key in IDENT_KEYS_ALL

    # -------------------------
    def _append_kv(self, key, val, kwargs, caller="append"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        if self.__is_identity_key(key):
            self.__add_identity(key, val, kwargs, caller=caller)
        else:
            is_k_valid = self.__validate_key_prefix(key, caller)
            if is_k_valid:
                self.__dict_append[key] = val

    def _remove_kv(self, key, val, kwargs, caller="remove"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        if self.__is_identity_key(key):
            self.__remove_identity(key, val, kwargs, caller=caller)
        else:
            is_k_valid = self.__validate_key_prefix(key, caller)
            if is_k_valid:
                self.__dict_remove[key] = val

    def _unset_k(self, key, caller="unset"):
        k, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        # ----
        self.__list_unset.append(k)

    def __add_identity(self, key, val, kwargs, caller):
        if key == IDENT_KEY_EMAIL:
            self._add_email(val, caller=caller)

        elif key == IDENT_KEY_SMS:
            self._add_sms(val, caller=caller)

        elif key == IDENT_KEY_WHATSAPP:
            self._add_whatsapp(val, caller=caller)

        elif key == IDENT_KEY_ANDROIDPUSH:
            self._add_androidpush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_append.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_append.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_IOSPUSH:
            self._add_iospush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_append.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_append.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_WEBPUSH:
            self._add_webpush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_append.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_append.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_SLACK:
            self._add_slack(val, caller=caller)

    def __remove_identity(self, key, value, kwargs, caller):
        if key == IDENT_KEY_EMAIL:
            self._remove_email(val, caller=caller)

        elif key == IDENT_KEY_SMS:
            self._remove_sms(val, caller=caller)

        elif key == IDENT_KEY_WHATSAPP:
            self._remove_whatsapp(val, caller=caller)

        elif key == IDENT_KEY_ANDROIDPUSH:
            self._remove_androidpush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_remove.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_remove.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_IOSPUSH:
            self._remove_iospush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_remove.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_remove.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_WEBPUSH:
            self._remove_webpush(val, kwargs.get(KEY_PUSHVENDOR), caller=caller)
            if self.__dict_remove.get(KEY_PUSHVENDOR):
                kwargs[KEY_PUSHVENDOR] = self.__dict_remove.get(KEY_PUSHVENDOR)

        elif key == IDENT_KEY_SLACK:
            self._remove_slack(val, caller=caller)

    # ------------------------
    def __check_ident_val_string(self, value, caller):
        msg = "value must a string with proper value"
        if not isinstance(value, (str,)):
            self.__errors.append(f"[{caller}] {msg}")
            return value, False
        value = value.strip()
        if not value:
            self.__errors.append(f"[{caller}] {msg}")
            return value, False
        # --
        return value, True

    # ------------------------ Email

    def __validate_email(self, email, caller):
        email, is_valid = self.__check_ident_val_string(email, caller)
        if not is_valid:
            return email, False
        # --- validate basic regex
        msg = "value in email format required. e.g. user@example.com"
        min_length, max_length = 6, 127
        # ---
        m = email_regex_compiled.match(email)
        if m is None:
            self.__errors.append(f"[{caller}] invalid value {email}. {msg}")
            return email, False
        if len(email) < min_length or len(email) > max_length:
            self.__errors.append(f"[{caller}] invalid value {email}. must be 6 <= len(email) <= 127")
            return email, False
        # ---
        return email, True

    def _add_email(self, value: str, caller: str):
        value, is_valid = self.__validate_email(value, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_EMAIL] = value

    def _remove_email(self, value: str, caller):
        value, is_valid = self.__validate_email(value, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_EMAIL] = value

    # ------------------------ Mobile no

    def __validate_mobile_no(self, mobile_no, caller):
        mobile_no, is_valid = self.__check_ident_val_string(mobile_no, caller)
        if not is_valid:
            return mobile_no, False
        # --- validate basic regex
        msg = "number must start with + and must contain country code. e.g. +41446681800"
        min_length = 8
        # ---
        m = mobile_regex_compiled.match(mobile_no)
        if m is None:
            self.__errors.append(f"[{caller}] invalid value {mobile_no}. {msg}")
            return mobile_no, False
        if len(mobile_no) < min_length:
            self.__errors.append(f"[{caller}] invalid value {mobile_no}. len(mobile_no) must be >= 8")
            return mobile_no, False
        # ---
        return mobile_no, True

    # ------------------------ SMS

    def _add_sms(self, value: str, caller: str):
        value, is_valid = self.__validate_mobile_no(value, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_SMS] = value

    def _remove_sms(self, value: str, caller: str):
        value, is_valid = self.__validate_mobile_no(value, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_SMS] = value

    # ------------------------ Whatsapp

    def _add_whatsapp(self, value: str, caller: str):
        value, is_valid = self.__validate_mobile_no(value, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_WHATSAPP] = value

    def _remove_whatsapp(self, value: str, caller: str):
        value, is_valid = self.__validate_mobile_no(value, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_WHATSAPP] = value

    # ------------------------ Androidpush [providers: fcm / xiaomi / oppo]

    def __check_androidpush_value(self, value, provider, caller):
        value, is_valid = self.__check_ident_val_string(value, caller)
        if not is_valid:
            return value, provider, False
        # -- validate provider
        if not provider:
            provider = "fcm"
        if provider not in ["fcm", "xiaomi", "oppo"]:
            self.__errors.append(f"[{caller}] unsupported androidpush provider {provider}")
            return value, provider, False
        # ---
        return value, provider, True

    def _add_androidpush(self, value: str, provider: str, caller: str):
        value, provider, is_valid = self.__check_androidpush_value(value, provider, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_ANDROIDPUSH] = value
        self.__dict_append[KEY_PUSHVENDOR] = provider

    def _remove_androidpush(self, value: str, provider: str, caller: str):
        value, provider, is_valid = self.__check_androidpush_value(value, provider, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_ANDROIDPUSH] = value
        self.__dict_remove[KEY_PUSHVENDOR] = provider

    # ------------------------ Iospush [providers: apns]

    def __check_iospush_value(self, value, provider, caller):
        value, is_valid = self.__check_ident_val_string(value, caller)
        if not is_valid:
            return value, provider, False
        # -- validate provider
        if not provider:
            provider = "apns"
        if provider not in ["apns", ]:
            self.__errors.append(f"[{caller}] unsupported iospush provider {provider}")
            return value, provider, False
        # ---
        return value, provider, True

    def _add_iospush(self, value: str, provider: str, caller: str):
        value, provider, is_valid = self.__check_iospush_value(value, provider, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_IOSPUSH] = value
        self.__dict_append[KEY_PUSHVENDOR] = provider

    def _remove_iospush(self, value: str, provider: str, caller: str):
        value, provider, is_valid = self.__check_iospush_value(value, provider, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_IOSPUSH] = value
        self.__dict_remove[KEY_PUSHVENDOR] = provider

    # ------------------------ Webpush [providers: vapid]

    def __check_webpush_dict(self, value, provider, caller):
        msg = "value must be a valid dict representing webpush-token"
        if not (value and isinstance(value, (dict,))):
            self.__errors.append(f"[{caller}] {msg}")
            return value, provider, False
        # -- validate provider
        if not provider:
            provider = "vapid"
        if provider not in ["vapid", ]:
            self.__errors.append(f"[{caller}] unsupported webpush provider {provider}")
            return value, provider, False
        # ---
        return value, provider, True

    def _add_webpush(self, value: dict, provider: str, caller: str):
        value, provider, is_valid = self.__check_webpush_dict(value, provider, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_WEBPUSH] = value
        self.__dict_append[KEY_PUSHVENDOR] = provider

    def _remove_webpush(self, value: dict, provider: str, caller: str):
        value, provider, is_valid = self.__check_webpush_dict(value, provider, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_WEBPUSH] = value
        self.__dict_remove[KEY_PUSHVENDOR] = provider

    # ------------------------ Slack

    def __validate_slack_userid(self, userid, caller):
        userid, is_valid = self.__check_ident_val_string(userid, caller)
        if not is_valid:
            return userid, False
        userid = userid.upper()
        if not (userid.startswith("U") or userid.startswith("W")):
            self.__errors.append(f"[{caller}] invalid value {userid}. Slack user/member_id starts with a U or W")
            return userid, False
        # -------
        return userid, True

    def __check_slack_dict(self, value, caller):
        msg = "value must be a valid dict/json with one of these keys: [email, user_id]"
        if not (value and isinstance(value, (dict,))):
            self.__errors.append(f"[{caller}] {msg}")
            return value, False
        user_id = value.get("user_id")
        email = value.get("email")
        if user_id and user_id.strip():
            user_id = user_id.strip()
            user_id, is_valid = self.__validate_slack_userid(user_id, caller)
            if not is_valid:
                return value, False
            else:
                return {"user_id": user_id}, True

        elif email and email.strip():
            email = email.strip()
            email, is_valid = self.__validate_email(email, caller)
            if not is_valid:
                return value, False
            else:
                return {"email": email}, True
        else:
            self.__errors.append(f"[{caller}] {msg}")
            return value, False

    def _add_slack(self, value: str, caller: str):
        value, is_valid = self.__check_slack_dict(value, caller)
        if not is_valid:
            return
        self.__dict_append[IDENT_KEY_SLACK] = value

    def _remove_slack(self, value: str, caller):
        value, is_valid = self.__check_slack_dict(value, caller)
        if not is_valid:
            return
        self.__dict_remove[IDENT_KEY_SLACK] = value



from datetime import datetime, timezone
import requests
import json
import copy
from typing import List, Dict

from .constants import (
    BODY_MAX_APPARENT_SIZE_IN_BYTES,
    BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
    MAX_EVENTS_IN_BULK_API,
    ALLOW_ATTACHMENTS_IN_BULK_API,
    HEADER_DATE_FMT,
)
from .signature import get_request_signature
from .bulk_response import BulkResponse
from .event import Event


class BulkEventsFactory:

    def __init__(self, config):
        self.config = config

    def new_instance(self):
        """
        USAGE:
        supr_client = Suprsend("__workspace_key__", "__workspace_secret__")
        bulk_ins = supr_client.bulk_events.new_instance()

        # append one by one
        for i in range(0, 10):
            event = Event('distinct_id', 'event_name', {}) # create event instance
            bulk_ins.append(event)

        # append many in one call
        all_events = [Event(..), Event(..), ...] # multiple events
        bulk_ins.append(*all_events)

        # call trigger
        response = bulk_ins.trigger()

        :return:
        """
        return BulkEvents(self.config)


class _BulkEventsChunk:
    _chunk_apparent_size_in_bytes = BODY_MAX_APPARENT_SIZE_IN_BYTES
    _chunk_apparent_size_in_bytes_readable = BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE
    _max_records_in_chunk = MAX_EVENTS_IN_BULK_API

    def __init__(self, config):
        self.config = config
        self.__chunk = []
        self.__url = self.__get_url()
        self.__headers = self.__common_headers()
        #
        self.__running_size = 0
        self.__running_length = 0
        self.response = None

    def __get_url(self):
        url_template = "{}event/"
        if self.config.include_signature_param:
            if self.config.auth_enabled:
                url_template = url_template + "?verify=true"
            else:
                url_template = url_template + "?verify=false"
        url_formatted = url_template.format(self.config.base_url)
        return url_formatted

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def __add_event_to_chunk(self, event, event_size):
        # First add size, then event to reduce effects of race condition
        self.__running_size += event_size
        self.__chunk.append(event)
        self.__running_length += 1

    def __check_limit_reached(self):
        if self.__running_length >= self._max_records_in_chunk or \
                self.__running_size >= self._chunk_apparent_size_in_bytes:
            return True
        else:
            return False

    def try_to_add_into_chunk(self, event: Dict, event_size: int) -> bool:
        """
        returns whether passed event was able to get added to this chunk or not,
        if true, event gets added to chunk
        :param event:
        :param event_size:
        :return:
        :raises: ValueError
        """
        if not event:
            return True
        if self.__check_limit_reached():
            return False
        # ---
        if event_size > self._chunk_apparent_size_in_bytes:
            raise ValueError(f"Event properties (discounting attachment if any) too big - {event_size} Bytes, "
                             f"must not cross {self._chunk_apparent_size_in_bytes_readable}")
        # if apparent_size of event crosses limit
        if self.__running_size + event_size > self._chunk_apparent_size_in_bytes:
            return False

        if not ALLOW_ATTACHMENTS_IN_BULK_API:
            event["properties"].pop("$attachments", None)

        # Add Event to chunk
        self.__add_event_to_chunk(event, event_size)
        return True

    def trigger(self):
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Based on whether signature is required or not, add Authorization header
        if self.config.auth_enabled:
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(self.__url, 'POST', self.__chunk, headers,
                                                     self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        else:
            content_txt = json.dumps(self.__chunk, ensure_ascii=False)
        # -----
        try:
            resp = requests.post(self.__url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            self.response = {
                "status": "fail",
                "status_code": 500,
                "total": len(self.__chunk),
                "success": 0,
                "failure": len(self.__chunk),
                "failed_records": [{"record": c, "error": error_str, "code": 500} for c in self.__chunk]
            }
        else:
            # TODO: handle 500/503 errors
            ok_response = resp.status_code // 100 == 2
            if ok_response:
                self.response = {
                    "status": "success",
                    "status_code": resp.status_code,
                    "total": len(self.__chunk),
                    "success": len(self.__chunk),
                    "failure": 0,
                    "failed_records": []
                }
            else:
                error_str = resp.text
                self.response = {
                    "status": "fail",
                    "status_code": resp.status_code,
                    "total": len(self.__chunk),
                    "success": 0,
                    "failure": len(self.__chunk),
                    "failed_records": [{"record": c, "error": error_str, "code": resp.status_code}
                                       for c in self.__chunk]
                }


class BulkEvents:
    def __init__(self, config):
        self.config = config
        self.__events = []
        self.__pending_records = []
        self.chunks = []
        self.response = BulkResponse()

    def __validate_events(self):
        if not self.__events:
            raise ValueError("events list is empty in bulk request")
        for ev in self.__events:
            ev_json, body_size = ev.get_final_json(self.config, is_part_of_bulk=True)
            self.__pending_records.append((ev_json, body_size))

    def __chunkify(self, start_idx=0):
        curr_chunk = _BulkEventsChunk(self.config)
        self.chunks.append(curr_chunk)
        for rel_idx, rec in enumerate(self.__pending_records[start_idx:]):
            is_added = curr_chunk.try_to_add_into_chunk(rec[0], rec[1])
            if not is_added:
                # create chunks from remaining records
                self.__chunkify(start_idx=(start_idx + rel_idx))
                # Don't forget to break. As current loop must not continue further
                break

    def append(self, *events):
        if not events:
            raise ValueError("events list empty. must pass one or more events")
        for ev in events:
            if not ev:
                raise ValueError("null/empty element found in bulk instance")
            if not isinstance(ev, Event):
                raise ValueError("element must be an instance of suprsend.Event")
            ev_copy = copy.deepcopy(ev)
            self.__events.append(ev_copy)

    def trigger(self):
        self.__validate_events()
        self.__chunkify()
        for c_idx, ch in enumerate(self.chunks):
            if self.config.req_log_level > 0:
                print(f"DEBUG: triggering api call for chunk: {c_idx}")
            # do api call
            ch.trigger()
            # merge response
            self.response.merge_chunk_response(ch.response)
        # -----
        return self.response



import json
import uuid
import time
from datetime import datetime, timezone
import requests
from typing import List, Dict

from .constants import (HEADER_DATE_FMT, BODY_MAX_APPARENT_SIZE_IN_BYTES, BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE)
from .attachment import get_attachment_json_for_file
from .signature import get_request_signature
from .utils import (validate_track_event_schema, get_apparent_event_size, )


RESERVED_EVENT_NAMES = [
    "$identify",
    "$notification_delivered", "$notification_dismiss", "$notification_clicked",
    "$app_launched", "$user_login", "$user_logout",
]


class Event:
    def __init__(self, distinct_id: str, event_name: str, properties: Dict = None):
        self.distinct_id = distinct_id
        self.event_name = event_name
        self.properties = properties
        # --- validate
        self.__validate_distinct_id()
        self.__validate_event_name()
        self.__validate_properties()

    def __validate_distinct_id(self):
        if not isinstance(self.distinct_id, (str,)):
            raise ValueError("distinct_id must be a string. an Id which uniquely identify a user in your app")
        distinct_id = self.distinct_id.strip()
        if not distinct_id:
            raise ValueError("distinct_id missing")
        self.distinct_id = distinct_id

    def __validate_properties(self):
        if self.properties is None:
            self.properties = {}
        if not isinstance(self.properties, (dict,)):
            raise ValueError("properties must be a dictionary")

    def __check_event_prefix(self, event_name: str):
        if event_name not in RESERVED_EVENT_NAMES:
            prefix_3_chars = event_name[:3].lower()
            if prefix_3_chars.startswith("$") or prefix_3_chars == "ss_":
                raise ValueError("event_names starting with [$,ss_] are reserved by SuprSend")

    def __validate_event_name(self):
        if not isinstance(self.event_name, (str,)):
            raise ValueError("event_name must be a string")
        event_name = self.event_name.strip()
        if not event_name:
            raise ValueError("event_name missing")
        self.__check_event_prefix(event_name)
        self.event_name = event_name

    def add_attachment(self, file_path: str):
        attachment = get_attachment_json_for_file(file_path)
        # --- add the attachment to body->data->$attachments
        if self.properties.get("$attachments") is None:
            self.properties["$attachments"] = []
        # -----
        self.properties["$attachments"].append(attachment)

    def get_final_json(self, config, is_part_of_bulk: bool = False):
        super_props = {"$ss_sdk_version": config.user_agent}
        event_dict = {
            "$insert_id": str(uuid.uuid4()),
            "$time": int(time.time() * 1000),
            "event": self.event_name,
            "env": config.workspace_key,
            "distinct_id": self.distinct_id,
            "properties": {**self.properties, **super_props}
        }
        event_dict = validate_track_event_schema(event_dict)
        # ---- Check size
        apparent_size = get_apparent_event_size(event_dict, is_part_of_bulk)
        if apparent_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise ValueError(f"Event properties too big - {apparent_size} Bytes, "
                             f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return event_dict, apparent_size


class EventCollector:
    def __init__(self, config):
        self.config = config
        self.__url = self.__get_url()
        self.__headers = self.__common_headers()

    def __get_url(self):
        url_template = "{}event/"
        if self.config.include_signature_param:
            if self.config.auth_enabled:
                url_template = url_template + "?verify=true"
            else:
                url_template = url_template + "?verify=false"
        url_formatted = url_template.format(self.config.base_url)
        return url_formatted

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def collect(self, event: Event) -> Dict:
        event_dict, event_size = event.get_final_json(self.config, is_part_of_bulk=False)
        return self.send(event_dict)

    def send(self, event: Dict) -> Dict:
        try:
            headers = {**self.__headers, **self.__dynamic_headers()}
            # Based on whether signature is required or not, add Authorization header
            if self.config.auth_enabled:
                # Signature and Authorization-header
                content_txt, sig = get_request_signature(self.__url, 'POST', event, headers,
                                                         self.config.workspace_secret)
                headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            else:
                content_txt = json.dumps(event, ensure_ascii=False)
            # -----
            resp = requests.post(self.__url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }



import platform

from typing import List, Dict
from .version import __version__
from .constants import (DEFAULT_URL, DEFAULT_UAT_URL)
from .exception import SuprsendConfigError
from .attachment import get_attachment_json_for_file
from .workflow import Workflow, _WorkflowTrigger
from .request_log import set_logging
from .workflows_bulk import BulkWorkflowsFactory
from .events_bulk import BulkEventsFactory
from .subscribers_bulk import BulkSubscribersFactory
from .subscriber import SubscriberFactory
from .event import Event, EventCollector


class Suprsend:
    """
    - Basic instance
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__")
    - Instance with debug on
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__", debug=True)
    - Instance with custom base-url
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__", base_url="https://example.com/", debug=False)
    """
    def __init__(self, workspace_key: str, workspace_secret: str, base_url: str = None, debug: bool = False, **kwargs):
        self.workspace_key = workspace_key
        self.workspace_secret = workspace_secret
        #
        self.sdk_version = __version__
        self.user_agent = "suprsend/{};python/{}".format(self.sdk_version, platform.python_version())
        #
        is_uat = kwargs.get("is_uat", False)
        self.base_url = self.__get_url(base_url, is_uat)
        # include cryptographic signature
        self.auth_enabled = (kwargs.get('auth_enabled') is not False)
        self.include_signature_param = (kwargs.get('include_signature_param') is not False)
        # ---
        self.__validate()
        # --- set logging level for http request
        self.req_log_level = 1 if debug else 0
        set_logging(self.req_log_level)
        #
        self._workflow_trigger = _WorkflowTrigger(self)
        self._eventcollector = EventCollector(self)
        # -- bulk instances
        self._bulk_workflows = BulkWorkflowsFactory(self)
        self._bulk_events = BulkEventsFactory(self)
        self._bulk_users = BulkSubscribersFactory(self)
        # --
        self._user = SubscriberFactory(self)

    @property
    def bulk_workflows(self):
        return self._bulk_workflows

    @property
    def bulk_events(self):
        return self._bulk_events

    @property
    def bulk_users(self):
        return self._bulk_users

    @property
    def user(self):
        return self._user

    @staticmethod
    def __get_url(base_url, is_uat):
        # ---- strip
        if base_url:
            base_url = base_url.strip()
        # ---- if url not passed, set url based on server env
        if not base_url:
            base_url = DEFAULT_UAT_URL if is_uat else DEFAULT_URL
        # ---- check url ends with /
        base_url = base_url.strip()
        if base_url[len(base_url) - 1] != "/":
            base_url = base_url + "/"
        return base_url

    def __validate(self):
        if not self.workspace_key:
            raise SuprsendConfigError("Missing workspace_key")
        if not self.workspace_secret:
            raise SuprsendConfigError("Missing workspace_secret")
        if not self.base_url:
            raise SuprsendConfigError("Missing base_url")

    def add_attachment(self, body: Dict, file_path: str) -> Dict:
        # if data key not present, add it and set value={}.
        if body.get("data") is None:
            body["data"] = {}
        if not isinstance(body, dict):
            raise ValueError("data must be a dictionary")
        # --------
        attachment = get_attachment_json_for_file(file_path)
        # --- add the attachment to body->data->$attachments
        if body["data"].get("$attachments") is None:
            body["data"]["$attachments"] = []
        #
        body["data"]["$attachments"].append(attachment)
        return body

    def trigger_workflow(self, data) -> Dict:
        """
        :param data:
        :return: {
            "success": true/false,
            "status": "success"/"fail",
            "message": "message",
            "status_code": 202/401/500,
        }
        :except:
            - SuprsendValidationError
        """
        if isinstance(data, Workflow):
            wf_ins = data
        else:
            wf_ins = Workflow(data)
        # -----
        return self._workflow_trigger.trigger(wf_ins)

    def track(self, distinct_id: str, event_name: str, properties: Dict = None) -> Dict:
        """
        :param distinct_id:
        :param event_name:
        :param properties:
        :return: {
            "success": True,
            "status": "success",
            "status_code": resp.status_code,
            "message": resp.text,
        }
        :except:
            - SuprsendValidationError (if post-data is invalid.)
            - ValueError
        """
        event = Event(distinct_id, event_name, properties)
        return self._eventcollector.collect(event)

    def track_event(self, event: Event) -> Dict:
        """
        :param event: suprsend.Event
        :return: {
            "success": True,
            "status": "success",
            "status_code": resp.status_code,
            "message": resp.text,
        }
        :except:
            - SuprsendValidationError (if post-data is invalid.)
            - ValueError
        """
        if not isinstance(event, Event):
            raise ValueError("argument must be an instance of suprsend.Event")
        return self._eventcollector.collect(event)
